
/*
 * LIB/PRECOMMIT.C	- Precommit caching
 *
 */

#include "defs.h"

Prototype int PreCommit(const char *msgid);
Prototype int SetPreCommitExpire(int pcex);

#define PC_HSIZE	4096	/* times sizeof(hash_t) */
#define PC_HMASK	(PC_HSIZE - 1)
#define PC_EXPIRE	30	/* 30 second expiration	*/

typedef struct pchash_t {
    hash_t	pc_Hash;
    time_t	pc_Time;
    pid_t	pc_Pid;
} pchash_t;

pchash_t *PCHAry;
int	PCFd = -1;
int	PCExpire = PC_EXPIRE;
int	PCPid = -1;

int 
SetPreCommitExpire(int pcex)
{
    int oex = PCExpire;
    PCExpire = pcex;
    return(oex);
}

int
PreCommit(const char *msgid)
{
    int r = 0;

    if (PCExpire == 0)
	return(0);

    if (PCPid == (pid_t)-1)
	PCPid = getpid();

    if (PCFd < 0) {
	struct stat st;

	PCFd = xopen(O_RDWR|O_CREAT, 0644, "%s/pcommit.cache", NewsHome);

	if (PCFd >= 0 && fstat(PCFd, &st) == 0) {
	    if (st.st_size < PC_HSIZE * sizeof(pchash_t))
		ftruncate(PCFd, PC_HSIZE * sizeof(pchash_t));
	    PCHAry = xmap(NULL, PC_HSIZE * sizeof(pchash_t), PROT_READ, MAP_SHARED, PCFd, 0);
	}
    }
    if (PCHAry == NULL && PCFd >= 0) {
	close(PCFd);
	PCFd = -1;
    }
    if (PCHAry) {
	hash_t hv = hhash(msgid);
	int i = (hv.h1 ^ hv.h2) & PC_HMASK;
	pchash_t *pc = &PCHAry[i];
	time_t t = time(NULL);
	long dt = t - pc->pc_Time;

	if (pc->pc_Hash.h1 == hv.h1 && 
	    pc->pc_Hash.h2 == hv.h2 &&
	    dt >= 0 && 
	    dt < PCExpire &&
	    pc->pc_Pid != PCPid
	) {
	    /*
	     * collision
	     */
	    r = -1;
	} else {
	    pchash_t npc;
	    /*
	     * enter new info, we don't care about collisions
	     */
	    npc.pc_Hash = hv;
	    npc.pc_Time = t - 1;
	    npc.pc_Pid = PCPid;
	    lseek(PCFd, i * sizeof(pchash_t), 0);
	    write(PCFd, &npc, sizeof(npc));
	}
    }
    return(r);
}

